/****************************************************************************/
/*                                                                          */
/*  This program using some code of FTView                                  */
/*	in ft2demo package of The FreeType project                              */
/*                                                                          */
/*  Copyright 2002 by                                                       */
/*  Dao Hai Lam <lam@visc-network.com>                                      */
/*                                                                          */
/*  FTCExtract - a simple character image extractor                         */
/*                                                                          */
/****************************************************************************/

#include "ftcommon.i"
#include FT_CACHE_MANAGER_H

#ifndef max
#define max(a,b) ((a) < (b) ? (b) : (a))
#endif

#include "utf.h"
#include "uni.dat"

int show_vn = 0;
int show_all = 0;

int startx = 300;
int starty = 80;
int image_height = 80;
int base_line = 16;
int char_width = 0;
int char_width_pad = 0;

char *folder = ".";

extern void png_save( char *fn, grBitmap *bit, int w );

static int fix_range(int *idx)
{
	if( !show_vn ) {
		if ( *idx < 32 ) {
			*idx = 32;
			return 1;
		}

		if ( *idx >= num_indices ) {
			*idx = num_indices - 1;
			return 1;
		}
	}
	else {
		if ( *idx < 0 ) {
			*idx = 0;
			return 1;
		}

		if ( !UTF16[*idx] ) {
			*idx -= 1;
			return 1;
		}
	}
	return 0;
}

static inline int get_index( int c )
{
	return show_vn ? get_glyph_index( UTF16[c] ) : get_glyph_index( c );
}

static FT_Error render_char(grBitmap *bit, int index, unsigned int code)
{
	FT_Pointer  glyf;
	grBitmap    bit3;
	int x_top, y_top;
	int left, top;
	int x_advance, y_advance;
	int sx, sy, stepy, n;

	error = FTC_Manager_Lookup_Size(cache_manager, &current_font.font,
			&face, &size );
	if ( error )
		return error;

	if ( code != 0 ) {
		sx = 0;
		sy = image_height - base_line;
	}
	else {
		sx = startx;
		sy = starty + (show_all ? 0 : (image_height - base_line));
		stepy = ( size->metrics.height >> 6 ) + 4;
		n = Num;
	}

__get_glyph:
	error = get_glyph_bitmap( index, &bit3, &left, &top,
			&x_advance, &y_advance, &glyf );

	if ( !error ) {
		int off;
		if ( code != 0 )
			off = max(left, 0);
		else
			off = left;
		x_top = sx + off;
		y_top = sy - top;
		grBlitGlyphToBitmap( bit, &bit3, x_top, y_top, fore_color );
		char_width = max(x_advance, bit3.width);
		if ( glyf ) done_glyph_bitmap( glyf );

		if ( !code && show_all ) {
			sx += x_advance + 1;
			if ( sx + size->metrics.x_ppem > bit->width ) {
				sx = startx;
				sy += stepy;
				if ( sy >= 240 )
					return FT_Err_Ok;
			}
		}
	}

	if ( !code && show_all ) {
		n++;
		if( !fix_range( &n ) ) {
			index = get_index( n );
			goto __get_glyph;
		}
	}
	return error;
}

static void save_char( grBitmap *b, unsigned int c )
{
	static unsigned char buffer[65536];
	int index = get_glyph_index( c );
	if ( !index ) return;

	b->buffer = buffer;
	memset(buffer, 0, b->width * b->rows);
	if ( render_char( b, index, c ) == 0 ) {
		static char s[1024];
		sprintf(s, "%s/%04x.png", folder, c);
		png_save(s, b, char_width);
	}
}

static void extract()
{
	unsigned int c;
	grBitmap b;

	b.mode = gr_pixel_mode_gray;
	b.rows = image_height;
	b.width = b.pitch = 300;
	b.grays = 256;

	for( c = 32; c < 256; c++ )
		save_char( &b, c );

	if ( !show_vn ) {
		for ( c = 256; c < num_indices; c++ )
			save_char( &b, c );
	}
	else {
		int n;
		for ( c = 0; n = UTF16[c]; c++ ) {
			if ( n<256 ) continue;
			save_char( &b, n );
		}
		for ( c = 0; n = extra_chars[c]; c++ ) {
			save_char( &b, n );
		}
	}
}

#define limit_range( value, vmin, vmax )	\
{											\
	if ( value < (vmin) )					\
		value = (vmin);						\
	else									\
	if ( value > (vmax) )					\
		value = (vmax);						\
}

static int process_event( grEvent*  event )
{
	int  i;

	switch ( event->key ) {
		case grKeyEsc:
		case grKEY( 'q' ):
		case grKEY( 'Q' ):
			return 0;

		case grKEY( 's' ):
		case grKEY( 'S' ):
			antialias  = !antialias;
			set_current_image_type();
			return 1;

		case grKEY( 'a' ):
		case grKEY( 'A' ):
			autohint = !autohint;
			set_current_image_type();
			return 1;
/*
		case grKEY( 'b' ): // toggle embedded bitmaps
			use_sbits  = !use_sbits;
			set_current_image_type();
			return 1;
*/
		case grKEY( 'h' ):
		case grKEY( 'H' ):
			hinted = !hinted;
			set_current_image_type();
			break;

		case grKEY( 'l' ):
		case grKEY( 'L' ):
			show_all ^= 1;
			if ( show_all ) {
				startx = 160;
				starty = ptsize;
			}
			else {
				startx = 300;
				starty = 80;
			}
			break;

		case grKEY( 'v' ):
		case grKEY( 'V' ):
			show_vn ^= 1;
			Num = show_vn ? 0 : 32;
			break;

		case grKEY( 'e' ):
		case grKEY( 'E' ):
			extract();
			break;

		case grKEY( 'p' ):
			if ( char_width_pad > 0 )
				char_width_pad--;
			break;
		case grKEY( 'P' ):
			if ( char_width_pad < 10 )
				char_width_pad++;
			break;

		case grKEY( '}' ):  i =   1; goto Do_ImageScale;
		case grKEY( '{' ):  i =  -1; goto Do_ImageScale;

		case grKEY( '>' ):  i =  10; goto Do_Scale;
		case grKEY( '<' ):  i = -10; goto Do_Scale;
		case grKEY( '+' ):  i =   1; goto Do_Scale;
		case grKEY( '-' ):  i =  -1; goto Do_Scale;

		case grKeyLeft:     i =  -1; goto Do_Glyph;
		case grKeyRight:    i =   1; goto Do_Glyph;

		case grKeyUp:       i =   1; goto Do_Baseline;
		case grKeyDown:     i =  -1; goto Do_Baseline;

		default:
			;
	}
	return 1;

Do_Baseline:
	if ( !show_all ) {
		base_line += i;
		limit_range( base_line, 0, image_height - ptsize );
	}
	return 1;

Do_Scale:
	ptsize += i;
	{
		int pts = ptsize;
		limit_range( ptsize, 1, image_height - base_line );
		if ( show_all && pts==ptsize )
			starty += i;
	}
	return 1;

Do_ImageScale:
	if ( !show_all ) {
		image_height += i;
		limit_range( image_height, ptsize, 144 );
		limit_range( base_line, 0, image_height - ptsize );
	}
	return 1;
	
Do_Glyph:
	Num += i;
	fix_range( &Num );

	return 1;
}

static void draw_frame(int x, int y, int w, int h)
{
	int offset = y * bit.width + x;
	int i, z = (h - 1) * bit.width;
	memset( &bit.buffer[offset], 0xFF, w );
	memset( &bit.buffer[offset+z], 0xFF, w );
	for ( i = 0; i < h; i++, offset += bit.width ) {
		bit.buffer[offset] =
		bit.buffer[offset + w - 1] = 0xFF;
	}
}

static void show_frame()
{
	draw_frame( 152, 4, 344, 256 );
	if ( !show_all ) {
		int x = startx - 1;
		int y = starty - 1;
		int w = char_width + char_width_pad;
		draw_frame( x, y, w + 2, image_height + 2 );
	}
}

static void show_string(char *fmt, ...)
{
	static char info[128];
	static int n = 0;
											
	if ( !fmt )
		n = 0;
	else {
		va_list ap;

		va_start(ap, fmt);
		vsnprintf(info, sizeof(info), fmt, ap);
		va_end(ap);
		grWriteCellString( &bit, 8, n++ * 16 + 8, info, fore_color );
	}
}

static void show_info()
{
	static char *status[] = { "off", "on" };
	static char others[256];

	show_string( 0 );
	show_string( "Resolution: %d", res);
	show_string( "[S]mooth: %s", status[antialias]);
	show_string( "[A]uto hint: %s", status[autohint]);
	show_string( "[H]int: %s", status[hinted]);
	show_string( "[L]ist: %s", status[show_all]);
	show_string( "Char: 0x%04x", show_vn ? UTF16[Num] : Num);
	show_string( "Size: %d", ptsize);
	show_string( "  +,-: inc/dec 1" );
	show_string( "  >,<: inc/dec 10" );
	show_string( "Image:" );
	show_string( "  Width: %d", char_width);
//	show_string( "    [P]ad: %d", char_width_pad);
	show_string( "  Height: %d", image_height);
	if( !show_all ) {
		show_string( "    },{: inc/dec" );
	}
	show_string( "  Base Line: %d", base_line);
	if( !show_all ) {
		show_string( "    [Up]: inc");
		show_string( "    [Dn]: dec");
	}

	sprintf( others, "[V]ietnamese character set: %s", status[show_vn]);
	grWriteCellString( &bit, 160, 250, "[<-] Prev                       Next [->]", fore_color );
	grWriteCellString( &bit, 160, 268, "[E]xtract characters to images", fore_color );
	grWriteCellString( &bit, 160, 284, others, fore_color );
	grWriteCellString( &bit, 448, 284, "[Q]uit", fore_color );
}

static void usage( char* execname )
{
	fprintf( stderr,  "\n" );
	fprintf( stderr,  "FTCExtract: simple character image extractor\n" );
	fprintf( stderr,  "-----------------------------------------------------------\n" );
	fprintf( stderr,  "\n" );
	fprintf( stderr,  "Usage: %s [options below] fontname[.ttf|.ttc] ...\n", execname );
	fprintf( stderr,  "\n" );
	fprintf( stderr,  "  -e            extract without display\n" );
	fprintf( stderr,  "  -v            Vietnamese characters set\n" );
	fprintf( stderr,  "  -s <points>   font size (default 64)\n" );
	fprintf( stderr,  "  -h <points>   char image height (default 80)\n" );
	fprintf( stderr,  "  -b <points>   base line position (default 16)\n" );
	fprintf( stderr,  "  -o <folder>   output image folder (default ./)\n" );
	fprintf( stderr,  "\n" );

	exit( 1 );
}

int main( int argc, char* argv[] )
{
	int     option;
	int     old_ptsize;
	int     orig_ptsize = 64;
	int     extract_only = 0;
	char*   execname;
	grEvent event;

	execname = ft_basename( argv[0] );

	while ( 1 ) {
		option = getopt( argc, argv, "evo:b:h:s:t:" );

		if ( option == -1 )
			break;

		switch ( option ) {
			case 'e':
				extract_only = 1;
				break;

			case 'v':
				show_vn = 1;
				break;

			case 'o':
				folder = optarg;
				break;

			case 'b':
				base_line = atoi( optarg );
				break;

			case 'h':
				image_height = atoi( optarg );
				if ( image_height < 24 )
					image_height = 24;
				break;

			case 's':
				orig_ptsize = atoi( optarg );
				if ( orig_ptsize < 24 )
					orig_ptsize = 24;
				break;

			case 't':
				hinted = atoi( optarg );
				break;

			default:
				usage( execname );
				break;
		}
	}

	argc -= optind;
	argv += optind;

	if ( argc < 1 )
		usage( execname );

	init_freetype();

	for ( ; argc > 0; argc--, argv++ )
		install_font_file( argv[0] );

	if ( num_fonts == 0 )
		PanicZ( "could not find/open any font file" );

	ptsize     = orig_ptsize;
	set_current_face( fonts[0] );
	set_current_pointsize( ptsize );
	set_current_image_type();
	num_indices = fonts[0]->num_indices;
	old_ptsize = ptsize;

	if ( orig_ptsize > image_height )
		orig_ptsize = image_height;
	if ( base_line > image_height - orig_ptsize )
		base_line = image_height - orig_ptsize;

	if( extract_only ) {
		extract();
		goto __done;
	}

	Init_Display();

	render_char( &bit, 0, 0 );
	if ( !face )
		grSetTitle( surface, "FT Character Extractor - Unknown" );
	else {
		static char title[256];
		sprintf( title, "FT Character Extractor 0.3 - %s (%s)",
				 face->family_name, face->style_name );
		grSetTitle( surface, title );
	}

	Num = 32;
	while ( 1 ) {
		Clear_Display();
		error = render_char( &bit, get_index( Num ), 0 );
		if ( error )
			fprintf( stderr, "Error: 0x%04x", (FT_UShort)error );

		show_info();
		show_frame();
		grRefreshSurface( surface );

		grListenSurface( surface, 0, &event );
		if ( !process_event( &event ) ) break;

		if ( ptsize != old_ptsize ) {
			set_current_pointsize( ptsize );
			old_ptsize = ptsize;
		}
	}

__done:
	done_freetype();
	exit( 0 );
	return 0;
}
