#include <stdio.h>
#include <png.h>
#include <graph.h>

#define uchar unsigned char

png_color palette[256];

void init_palette()
{
	int i;
	static int ready = 0;
	if( ready ) return;

	ready = 1;
	for( i=0; i<256; i++ ) {
		png_color c = {(uchar)i, (uchar)i, (uchar)i};
		palette[i] = c;
	}

}

void png_my_error( png_structp png_ptr, png_const_charp message )
{                                                                                                    
	fprintf( stderr, "ERROR(libpng): %s - %s\n",
				message, (char *)png_get_error_ptr(png_ptr) );
	longjmp( png_ptr->jmpbuf, 1 );
}

void png_my_warning( png_structp png_ptr, png_const_charp message )
{
	fprintf( stderr, "WARNING(libpng): %s - %s\n",
			message, (char *)png_get_error_ptr(png_ptr) );
}

void png_save( char *fn, grBitmap *bit, int w )
{
	static uchar *buffer = NULL;
	static png_bytep b[256];

	FILE *fp = fopen(fn, "w");
	if( !fp ) return;

	if( buffer != bit->buffer ) {
		uchar *bp = bit->buffer;
		int i;

		buffer = bp;
		for( i=0; i<256; i++, bp+=bit->width )
			b[i] = (png_bytep)bp;
	}
	png_write( fn, fp, b, w, bit->rows );
	fclose( fp );
}

int png_write( char *fn, FILE *fp, png_bytepp buf, int width, int height )
{
	png_structp png_ptr;
	png_infop info_ptr;

	png_ptr = png_create_write_struct( PNG_LIBPNG_VER_STRING, fn,
										png_my_error, png_my_warning );
	if( !png_ptr )
		return 0;

	info_ptr = png_create_info_struct( png_ptr );
	if( !info_ptr ) {
		png_destroy_write_struct( &png_ptr, NULL );
		return 0;
	}

	if( setjmp(png_ptr->jmpbuf) ) {
		png_destroy_write_struct( &png_ptr, &info_ptr );
		return 0;
	}

	init_palette();
	png_init_io( png_ptr, fp );
	png_set_compression_level( png_ptr, 9 );
	png_set_filter( png_ptr, PNG_FILTER_TYPE_BASE, PNG_NO_FILTERS);

//	png_set_PLTE( png_ptr, info_ptr, palette, 256 );
	png_set_IHDR( png_ptr, info_ptr, width, height, 8,
					PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE,
					PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT );
	png_write_info( png_ptr, info_ptr );
	png_write_image( png_ptr, buf );
	png_write_end( png_ptr, info_ptr );
	png_destroy_write_struct( &png_ptr, &info_ptr );

	return 1;
}

/*
int main( int argc, char *argv[] )
{
	static png_byte data[65536];
	FILE *fp = fopen("a.png", "w");
	png_bytep buf[100];
	int i;

	if( !fp ) {
		fprintf( stderr, "Create file error!" );
		return -1;
	}

	for( i=0; i<256; i++ ) {
		png_color c = {(unsigned char)i, (unsigned char)i, (unsigned char)i};
		palette[i] = c;
	}

	for( i=0; i<100; i++ )
		buf[i] = &data[i*37];

	memset(&buf[1][1], 0xFF, 35);
	memset(&buf[98][1], 0xFF, 35);
	for( i=2; i<98; i++ )
		buf[i][1] = buf[i][35] = 0xFF;

	png_save( "a.png", fp, buf, 37, 100 );
	fclose( fp );
	return 0;
}
*/
